#!/usr/bin/env node

let LineBuffer = require('./util').LineBuffer;
let args = require('minimist')(process.argv.slice(2));
let host = args.host || "localhost";
let port = parseInt(args.port || 3000);
let user = args.user || undefined;
let pass = args.pass || undefined;
let base = args.base || undefined;
let client = new (require('./index').client)();
let verbose = !args.quiet;

let exiting = false;
let limit = undefined;
let promise;

if (verbose) {
    console.log("[ grid.space net level client 1.0 ]");
}

function onerror(error) {
    console.log({db_error: error});
}

if (host) {
    promise = client.open(host, port).catch(onerror);
    if (user) {
        promise = promise.then(() => {
            return client.auth(user,pass);
        }).catch(onerror);
    }
    if (base) {
        promise = promise.then(() => {
            return client.use(base);
        }).catch(onerror);
    }
}

function prompt() {
    if (exiting) {
        process.exit();
    }
    if (verbose) {
        process.stdout.write(`:: `);
    }
}

function error(scope) {
    return function(error) {
        console.log({[`${scope}_error`]: error});
        prompt();
    }
}

if (promise) {
    promise.then(prompt).then(start);
} else {
    prompt();
    start();
}

function start() {
    new LineBuffer(process.stdin, next);
    process.stdin.on('close', () => {
        exiting = true;
    });
}

function next(line) {
    line = line.toString().trim().split(' ').map(v => v.trim());
    let cmd = line.shift();
    switch (cmd) {
        case 'halt':
            client.halt().then(prompt).catch(error('halt'));
            break;
        case 'drop':
            client.drop(line[0]).catch(error('drop'));
            prompt();
            break;
        case 'debug':
            let value;
            try {
                value = eval(`(${line[1]})`);
            } catch (err) {
                console.log(err);
                prompt();
                break;
            }
            client.debug(line[0], value).then(prompt).catch(error('debug'));
            break;
        case 'close':
            client.close();
            prompt();
            break;
        case 'open':
            client.close()
                .open(line[0] || client.host || host, parseInt(line[1] || client.port || port))
                .then(prompt).catch(error('open'));
            break;
        case 'stat':
            client.stat({[line[0]] : true}).then(reply => {
                let stat = reply.stat;
                if (Array.isArray(stat)) {
                    stat = stat.join("\n");
                }
                if (stat) {
                    console.log(stat);
                }
                prompt();
            }).catch(error('stat'));
            break;
        case 'auth':
            client.auth(line[0],line[1]).then(prompt).catch(error('auth'));
            break;
        case 'use':
            let dbn = line[0] || [42];
            let enc = line[1] ? { valueEncoding: line[1] === 'none' ? undefined : line[1] } : undefined;
            client.use(dbn, enc).then(prompt).catch(error('use'));
            break;
        case 'get':
            client.get(line[0]).then(reply => {
                console.log(reply);
                prompt();
            }).catch(error('get'));
            break;
        case 'put':
            let [ key, val ] = line;
            try {
                val = eval(`(${val})`);
            } catch (err) {
                console.log(err);
                prompt();
                break;
            }
            client.put(key, val).then(reply => {
                prompt();
            }).catch(error('put'));
            break;
        case 'del':
            client.del(line[0]).then(prompt).catch(error('del'));
            break;
        case 'limit':
            limit = line[0] ? parseInt(line[0]) : undefined;
            prompt();
            break;
        case 'ls':
            let ls = {};
        case 'count':
        case 'list':
        case 'keys':
            let isLS = cmd === 'ls';
            let isCount = cmd === 'count';
            let isKeys = cmd === 'keys' || isCount || isLS;
            let opt = line.length === 1 || isLS ? {
                pre: line[0]
            } : {
                gte: line[0],
                lt: line[1]
            }
            let count = 0;
            client.list(Object.assign({
                values: isKeys ? false : true,
                limit
            }, opt), (key, value) => {
                count++;
                if (!isCount) {
                    if (isLS) {
                        let kv = (line[0] ? key.substring(line[0].length) : key).split('/')[0];
                        if (kv.length === 0) {
                            return;
                        }
                        if (!ls[kv]) {
                            console.log(kv);
                            ls[kv] = 1;
                        } else {
                            ls[kv]++;
                        }
                    } else {
                        console.log(isKeys ? key : {key, value});
                    }
                }
            }).then(() => {
                if (isLS) {
                    for (let key in ls) {
                        if (ls[key] === 1) {
                            delete ls[key];
                        }
                    }
                    if (Object.keys(ls).length) {
                        console.log(ls);
                    }
                }
                console.log({count});
                prompt();
            }).catch(error('list'));
            break;
        case '':
            prompt();
            break;
        default:
            console.log({invalid_command: cmd});
            prompt();
            break;
    }
}
