#!/usr/bin/env node

let fs = require('fs');
let net = require('net');
let level = require('level');
let { LineBuffer, encode, decode, lastmod, log } = require('./util');

function init(args) {
    let uname = args.user || 'user';
    let upass = args.pass || Date.now().toString(36) + (Math.random() * 0xffffff).toString(36);
    let port = parseInt(args.port || 3000);
    let ddir = args.dir || args.base || 'data';
    let nuid = Date.now();
    let base = {};
    let dbug = {};
    let user = {
        [uname]: {
            pass: args.pass || upass,
            perms: {
                halt: true,
                stat: true,
                drop: true,
                create: true,
                put: true,
                del: true,
                range: Infinity
            }
        }
    };

    log(`starting net-level on port ${port} serving "${ddir}"`);

    let server = net.createServer(conn => {
        let addr = conn.remoteAddress;
        let gets = 0;
        let puts = 0;
        let iter = 0;
        let mark = Date.now();
        let uuid = nuid++;
        let unam = 'unknown';
        let name;
        let rec;
        let db;
        let defperms = {
            halt: false,
            stat: false,
            drop: false,
            create: false,
            put: false,
            del: false,
            range: 10
        };
        let perms = defperms;

        function close(rec) {
            if (rec) {
                let uids = rec.uids;
                let pos = uids.indexOf(uuid);
                if (pos >= 0) {
                    uids.splice(pos,1);
                }
                if (uids.length === 0) {
                    if (db) {
                        db.close();
                    }
                    db = undefined;
                    delete base[name];
                }
            }
        }

        function check(call) {
            if (!db) {
                conn.write(encode({call, error: "no database in use"}));
                return false;
            }
            return true;
        }

        function deleteDir(dir, then) {
            fs.readdir(dir, (error, files) => {
                deleteFiles(dir, files, then);
            });
        }

        function deleteFiles(dir, files, then) {
            if (files.length) {
                let file = files.shift();
                fs.unlink(`${dir}/${file}`, error => {
                    if (error) {
                        then(error);
                    } else {
                        deleteFiles(dir, files, then);
                    }
                });
            } else {
                fs.rmdir(dir, error => {
                    then(error);
                });
            }
        }

        new LineBuffer(conn, line => {
            let json = decode(line);
            let { call, stat, auth, pass, halt, drop, debug } = json;
            let { use, get, put, del, list, opt, value } = json;

            if (halt && perms.halt) {
                server.close();
                for (let rec of Object.values(base)) {
                    close(rec);
                }
                conn.write(encode({call, halt}));
                process.exit();
            } else if (stat && perms.stat) {
                fs.readdir(ddir, (err, list) => {
                    stat = {
                        open: Object.keys(base),
                        list,
                        mark,
                        gets,
                        puts,
                        iter,
                        dbug
                    };
                    if (opt) {
                        if (opt.list) {
                            stat = stat.list
                        } else if (opt.open) {
                            stat = stat.open;
                        }
                    }
                    conn.write(encode({call, stat}));
                });
            } else if (drop && perms.drop) {
                let dropdir = `${ddir}/${drop}`;
                if (base[drop]) {
                    conn.write(encode({call, drop, error: "database in use"}));
                } else if (lastmod(dropdir)) {
                    deleteDir(dropdir, error => {
                        conn.write(encode({call, drop, error}));
                    });
                } else {
                    conn.write(encode({call, drop, error: "no such database"}));
                }
            } else if (auth) {
                let urec = user[auth];
                if (urec && urec.pass === pass) {
                    perms = urec.perms;
                    log({addr, auth, perms});
                } else {
                    perms = defperms;
                }
                unam = auth;
                conn.write(encode({call, auth}));
            } else if (use) {
                close(rec);
                if (Array.isArray(use) && use.length === 1 && use[0] === 42) {
                    rec = undefined;
                    return conn.write(encode({call, use}));
                }
                name = use;
                rec = base[use];
                let path = `${ddir}/${use}`;
                if (rec) {
                    db = rec.db;
                    if (rec.uids.indexOf(uuid) < 0) {
                        rec.uids.push(uuid);
                    }
                    conn.write(encode({call, use}));
                    log({addr, user: unam, open: path});
                } else if (lastmod(path) || perms.create) {
                    db = level(path, opt || { valueEncoding: "json" });
                    db.on('error', error => {
                        console.log({db_error: error});
                        close(base[use]);
                    });
                    rec = base[use] = { uids: [ uuid ], db };
                    conn.write(encode({call, use}));
                    log({addr, user: unam, open: path});
                } else {
                    conn.write(encode({call, use, error: "not authorized"}));
                }
            } else if (get) {
                gets++;
                if (check(call))
                db.get(get, (error, value) => {
                    if (error && error.notFound) {
                        error = undefined;
                        value = undefined;
                    }
                    conn.write(encode({call, get, value, error}));
                    if (dbug[name]) {
                        log({addr, user: unam, get, error});
                    }
                });
            } else if (put) {
                if (!perms.put) {
                    return conn.write(encode({call, put, error: "not authorized"}));
                }
                puts++;
                if (check(call))
                db.put(put, value, error => {
                    conn.write(encode({call, put, error}));
                    if (dbug[name]) {
                        log({addr, user: unam, put, error});
                    }
                });
            } else if (del) {
                if (!perms.del) {
                    return conn.write(encode({call, put, error: "not authorized"}));
                }
                if (check(call))
                db.del(del, error => {
                    conn.write(encode({call, del, error}));
                    if (dbug[name]) {
                        log({addr, user: unam, del, error});
                    }
                });
            } else if (list) {
                let read = 0;
                if (!check(call)) {
                    return;
                }
                iter++;
                let pre = opt.pre;
                if (pre) {
                    opt.gte = pre;
                }
                let stream = db.createReadStream(opt)
                    .on('error', error => {
                        conn.write(encode({call, error}));
                    })
                    .on('data', data => {
                        if (pre) {
                            let key = data.key || data;
                            if (!key.startsWith(pre)) {
                                conn.write(encode({call}));
                                return stream.destroy();
                            }
                        }
                        let cont = read++ < perms.range;
                        conn.write(encode({call, list: data, continue: cont}));
                        if (!cont) {
                            stream.destroy();
                        }
                    })
                    .on('end', () => {
                        conn.write(encode({call}));
                        if (dbug[name]) {
                            log({addr, user: unam, list, read});
                        }
                    });
            } else if (debug) {
                dbug[debug] = value;
                conn.write(encode({call, debug}));
            } else {
                conn.write(encode({call, error: "no command"}));
            }
        });
        conn.on('close', () => {
            close(rec);
        });
        conn.on('error', error => {
            log({error});
        });
    }).listen(port);
}

if (module.parent) {
    module.exports = init;
} else {
    init(require('minimist')(process.argv.slice(2)));
}
